/*
 * Copyright (c) 1998-2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 * 
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */

#include <IOKit/firewire/IOFireWireBus.h>
#include <IOKit/firewire/IOFireWireNub.h>
#include <IOKit/IOSyncer.h>
 
#include "IOFWAsyncStreamRxCommand.h"


OSDefineMetaClassAndStructors(IOFWAsyncStreamRxCommand, OSObject);
OSMetaClassDefineReservedUnused(IOFWAsyncStreamRxCommand, 0);
OSMetaClassDefineReservedUnused(IOFWAsyncStreamRxCommand, 1);

/*!
    @function initAll
	Initializes the AsyncStream Recieve command object
	@result true if successfull.
*/
bool IOFWAsyncStreamRxCommand::initAll(UInt32 channel, DCLCallCommandProc* proc, 
									   IOFireWireController *control,
									   void	*callbackObject)
{    
	IOReturn status	= kIOReturnSuccess;
	
	fControl = control;
	fFWIM = fControl->getLink();
	
	
	// Create a DCL program
	fdclProgram = CreateAsyncStreamRxDCLProgram(proc, callbackObject);
	if(fdclProgram == NULL){
		IOLog("    Error: AddAsyncStreamClient %d %x\n", __LINE__, status);
		return false;
	}

	fChan = channel;
	fSpeed = kFWSpeedMaximum;

	bStarted = false;
    fInitialized = true;
    
	// Start the channel
	// start();

	return true;
}

void IOFWAsyncStreamRxCommand::free()
{
	// bStarted = true;

	//IOLog("    AddAsyncStreamClient Release called %d\n", __LINE__);
	
	stop();

	// free dcl program
	FreeAsyncStreamRxDCLProgram(fdclProgram);

	// FIREWIRETODO : Should I call before port free ??
	// free the IODCLProgram
	// fIODclProgram->release();
	
	// free the buffer 
	fBufDesc->release();
	
	OSObject::free();
}

void IOFWAsyncStreamRxCommand::FreeAsyncStreamRxDCLProgram(DCLCommandStruct *dclProgram) {

	DCLLabel			*startLabel;
	DCLTransferPacket	*receivePacket;
	DCLUpdateDCLList	*update;
	DCLCommand			*dclCommand;
	DCLCallProc			*callProc;
	RXProcData			*rxProcData;

	if(dclProgram == NULL)
		return;
		
	startLabel = (DCLLabel*)dclProgram;
	if(startLabel == NULL)
		return;
		
	receivePacket = (DCLTransferPacket*)startLabel->pNextDCLCommand;
	if(receivePacket == NULL)
		return;
	
	update = (DCLUpdateDCLList*)receivePacket->pNextDCLCommand;
	if(update == NULL)
		return;
	
	dclCommand = (DCLCommand*)update->dclCommandList;
	if(dclCommand == NULL)
		return;
	
	callProc = (DCLCallProc*)update->pNextDCLCommand;
	if(callProc == NULL)
		return;
	
	rxProcData = (RXProcData*)callProc->procData;
	if(rxProcData == NULL)
		return;
	
	delete rxProcData;
	delete dclCommand;
	delete update;
	delete receivePacket;
	delete startLabel;
		
	DCLCallProc			*lastBlockLastProc = callProc;
	receivePacket = (DCLTransferPacket*)lastBlockLastProc->pNextDCLCommand;
	update = (DCLUpdateDCLList*)receivePacket->pNextDCLCommand;
	dclCommand = (DCLCommand*)update->dclCommandList;
	callProc = (DCLCallProc*)update->pNextDCLCommand;
	DCLJump				*jump = (DCLJump *)callProc->pNextDCLCommand;
	rxProcData = (RXProcData*)callProc->procData;
	
	delete rxProcData;
	delete jump;
	delete dclCommand;
	delete update;
	delete receivePacket;
	delete lastBlockLastProc;
	delete callProc;
}

DCLCommandStruct *IOFWAsyncStreamRxCommand::CreateAsyncStreamRxDCLProgram(DCLCallCommandProc* proc, void *callbackObject) {
	
	// Create a DCL program
	fBufDesc = new IOBufferMemoryDescriptor;
    if(fBufDesc == NULL)
        return NULL;

	if(!fBufDesc->initWithOptions( 0, 2 * kMaxIsochPacketSize, PAGE_SIZE))	// get buffer large enough for 2 packets
	{
        return NULL;
	}
    
	UInt8*				currentBuffer = (UInt8*)fBufDesc->getBytesNoCopy() ;
	DCLLabel*			startLabel = new DCLLabel ;
	{
		startLabel->opcode = kDCLLabelOp ;
	}
	
	DCLCallProc*			lastBlockLastProc = NULL ;

	// 1st buffer receive
	{
		DCLTransferPacket*	receivePacket = new DCLTransferPacket ;
		{
			receivePacket->opcode = kDCLReceivePacketStartOp ;
			receivePacket->buffer = currentBuffer ;
			receivePacket->size = kMaxIsochPacketSize ;
		}
		
		DCLUpdateDCLList*	update = new DCLUpdateDCLList ;
		{
			update->opcode = kDCLUpdateDCLListOp ;
			update->dclCommandList = new DCLCommand*[1] ;
			update->dclCommandList[0] = (DCLCommand*)receivePacket ;
			update->numDCLCommands = 1;		// Number of DCL commands in list.
		}
	
		DCLCallProc*		callProc = new DCLCallProc ;
		{
			callProc->opcode = kDCLCallProcOp ;
			callProc->proc = proc ;
			callProc->procData = (UInt32)new RXProcData ;
			
			((RXProcData*)callProc->procData)->obj = callbackObject;
			((RXProcData*)callProc->procData)->buffer = currentBuffer ;
		}

		startLabel->pNextDCLCommand = (DCLCommand*)receivePacket ;
		receivePacket->pNextDCLCommand = (DCLCommand*)update ;
		update->pNextDCLCommand = (DCLCommand*)callProc ;

		lastBlockLastProc = callProc ;

		currentBuffer += receivePacket->size ;
	}
	
	// 2nd buffer receive
	{
		DCLTransferPacket*	receivePacket = new DCLTransferPacket ;
		{
			receivePacket->opcode = kDCLReceivePacketStartOp ;
			receivePacket->buffer = currentBuffer ;
			receivePacket->size = kMaxIsochPacketSize ;
		}
		
		DCLUpdateDCLList*	update = new DCLUpdateDCLList ;
		{
			update->opcode = kDCLUpdateDCLListOp ;
			update->dclCommandList = new DCLCommand*[1] ;
			update->dclCommandList[0] = (DCLCommand*)receivePacket ;
			update->numDCLCommands = 1;		// Number of DCL commands in list.
		}
	
		DCLCallProc*		callProc = new DCLCallProc ;
		{
			callProc->opcode = kDCLCallProcOp ;
			callProc->proc = proc;
			callProc->procData = (UInt32)new RXProcData ;
			
			((RXProcData*)callProc->procData)->obj = callbackObject;
			((RXProcData*)callProc->procData)->buffer = currentBuffer ;
		}

		DCLJump*	jump = new DCLJump ;
		{
			jump->opcode = kDCLJumpOp ;
			jump->pJumpDCLLabel = startLabel ;
		}
		
		// link up allocated DCLs
		lastBlockLastProc->pNextDCLCommand = (DCLCommand*)receivePacket ;
		receivePacket->pNextDCLCommand = (DCLCommand*)update ;
		update->pNextDCLCommand = (DCLCommand*)callProc ;
		callProc->pNextDCLCommand = (DCLCommand*)jump ;
		jump->pNextDCLCommand = NULL ;
	}
	
	return (DCLCommand*)startLabel ;
}

IOIPPort *IOFWAsyncStreamRxCommand::CreateAsyncStreamPort(bool talking, DCLCommandStruct *opcodes, void *info,
													UInt32 startEvent, UInt32 startState, UInt32 startMask,
													UInt32 channel )
{
    IOIPPort *port;

    if(fFWIM == NULL)
    {
		IOLog("    %s:%d fFWIM not initialized\n", __FILE__, __LINE__);
		return NULL;
    }

    fIODclProgram = fFWIM->createDCLProgram(talking, opcodes, NULL, startEvent, startState, startMask);
    if(!fIODclProgram) 
	{
		IOLog("    %s:%d IODclProgram returned NULL from FWIM\n", __FILE__, __LINE__);
		return NULL;
	}

    port = new IOIPPort;
    if(!port) {
		fIODclProgram->release();
		return NULL;
    }

    if(!port->init(fIODclProgram, fControl, channel)) {
		IOLog("    %s:%d IOIPPort init failed \n", __FILE__, __LINE__);
		port->release();
		port = NULL;
    }

    return port;
}

IOReturn IOFWAsyncStreamRxCommand::start(IOFWSpeed fBroadCastSpeed) {

	IOReturn status	= kIOReturnSuccess;

    if(fInitialized == false)
        return status;

	if(bStarted) 
		return status;

	// IOLog("    starting channel ... %d\n", __LINE__);

	// Create IOIPPort with the required channel and DCL program
	fAsynStreamPort = CreateAsyncStreamPort(false, fdclProgram, NULL, 0, 0, 0, fChan);
	if(fAsynStreamPort == NULL){
		FreeAsyncStreamRxDCLProgram(fdclProgram);
		IOLog("    Error: AddAsyncStreamClient %d %x\n", __LINE__, status);
		return kIOReturnError;
	}

	fSpeed = fBroadCastSpeed;

	// Create a IOFWIsocChannel with the created port
	fAsyncStreamChan = fControl->createIsochChannel(false, 0, fSpeed, NULL, NULL);

	if(fAsyncStreamChan == NULL) {
		FreeAsyncStreamRxDCLProgram(fdclProgram);
		fAsynStreamPort->release() ;
		IOLog("    Error: AddAsyncStreamClient %d %x\n", __LINE__, status);
		return kIOReturnError;
	}
	
	status = fAsyncStreamChan->addListener(fAsynStreamPort);
	if(status != kIOReturnSuccess){
		FreeAsyncStreamRxDCLProgram(fdclProgram);
		fAsyncStreamChan->releaseChannel();
		fAsynStreamPort->release() ;
		IOLog("    Error: AddAsyncStreamClient %d %x\n", __LINE__, status);
		return kIOReturnError;
	}
	
	// Allocate channel
	status = fAsyncStreamChan->allocateChannel();
#ifdef FIREWIRETODO // Ignore the error right now
	if(status != kIOReturnSuccess){
		FreeAsyncStreamRxDCLProgram(fdclProgram);
		fAsyncStreamChan->releaseChannel();
		fAsynStreamPort->release() ;
		IOLog("    Error: AddAsyncStreamClient %d %x\n", __LINE__, status);
		return kIOReturnError;
	}
#endif
	
	// Start channel
	status = fAsyncStreamChan->start();
	if(status != kIOReturnSuccess){
		FreeAsyncStreamRxDCLProgram(fdclProgram);
		fAsyncStreamChan->releaseChannel();
		fAsynStreamPort->release() ;
		IOLog("    Error: AddAsyncStreamClient %d %x\n", __LINE__, status);
		return kIOReturnError;
	}
	
	bStarted = true;
		
	return status;
}
	
IOReturn IOFWAsyncStreamRxCommand::stop() {
	IOReturn status	= kIOReturnSuccess;

    if(fInitialized == false)
        return status;

	if(!bStarted) 
		return status;

	// IOLog("    stopping channel ... %d\n", __LINE__);
	
	// stop receiving packets
	fAsynStreamPort->stop();
	
	// free the port
	fAsynStreamPort->release();
	
	// Stop the channel
	fAsyncStreamChan->stop();

	// Lets release the channel
	fAsyncStreamChan->releaseChannel();
	
	// free the channel	
	fAsyncStreamChan->release();
	
	bStarted = false;
	
	return status;
}


                                                                                                    
